@@ -114,6 +114,7 @@ def get_distutils_distro(_cache=[]):
 
 EXT_DEP_MODULES = {
     'tag:numpy':    'numpy',
+    'tag:pythran':  'pythran',
     'tag:asyncio':  'asyncio',
     'tag:pstats':   'pstats',
     'tag:posix':    'posix',
@@ -121,7 +122,7 @@ EXT_DEP_MODULES = {
     'tag:coverage': 'Cython.Coverage',
     'Coverage':     'Cython.Coverage',
     'tag:ipython':  'IPython',
-    'tag:jedi':     'jedi',
+    'tag:jedi':     'jedi_BROKEN_AND_DISABLED',
 }
 
 def patch_inspect_isfunction():
@@ -315,10 +316,6 @@ EXT_EXTRAS = {
 }
 
 
-def _is_py3_before_32(excluded, version):
-    return version[0] >= 3 and version < (3,2)
-
-
 # TODO: use tags
 VER_DEP_MODULES = {
     # tests are excluded if 'CurrentPythonVersion OP VersionTuple', i.e.
@@ -341,14 +338,16 @@ VER_DEP_MODULES = {
                                         'compile.extdelslice',
                                         'run.special_methods_T561_py2'
                                         ]),
-    (3,1): (_is_py3_before_32, lambda x: x in ['run.pyclass_special_methods',
-                                               ]),
     (3,3) : (operator.lt, lambda x: x in ['build.package_compilation',
                                           'run.yield_from_py33',
                                           ]),
     (3,4): (operator.lt, lambda x: x in ['run.py34_signature',
                                          ]),
+    (3,4,999): (operator.gt, lambda x: x in ['run.initial_file_path',
+                                             ]),
     (3,5): (operator.lt, lambda x: x in ['run.py35_pep492_interop',
+                                         'run.mod__spec__',
+                                         'run.pep526_variable_annotations',  # typing module
                                          ]),
 }
 
@@ -581,9 +580,16 @@ class TestBuilder(object):
             for test in self.build_tests(test_class, path, workdir,
                                          module, mode == 'error', tags):
                 suite.addTest(test)
+
             if mode == 'run' and ext == '.py' and not self.cython_only and not filename.startswith('test_'):
                 # additionally test file in real Python
-                suite.addTest(PureDoctestTestCase(module, os.path.join(path, filename)))
+                min_py_ver = [
+                    (int(pyver.group(1)), int(pyver.group(2)))
+                    for pyver in map(re.compile(r'pure([0-9]+)[.]([0-9]+)').match, tags['tag'])
+                    if pyver
+                ]
+                if not min_py_ver or any(sys.version_info >= min_ver for min_ver in min_py_ver):
+                    suite.addTest(PureDoctestTestCase(module, os.path.join(path, filename)))
 
         return suite
 
@@ -606,10 +612,16 @@ class TestBuilder(object):
             languages = list(languages)
             languages.remove('cpp')
 
+        pythran_dir = self.pythran_dir
+        if 'pythran' in tags['tag'] and not pythran_dir and 'cpp' in languages:
+            import pythran.config
+            pythran_ext = pythran.config.make_extension()
+            pythran_dir = pythran_ext['include_dirs'][0]
+
         preparse_list = tags.get('preparse', ['id'])
         tests = [ self.build_test(test_class, path, workdir, module, tags, language,
                                   expect_errors, expect_warnings, warning_errors, preparse,
-                                  self.pythran_dir if language == "cpp" else None)
+                                  pythran_dir if language == "cpp" else None)
                   for language in languages
                   for preparse in preparse_list ]
         return tests
@@ -880,9 +892,6 @@ class CythonCompileTestCase(unittest.TestCase):
             if extra_extension_args is None:
                 extra_extension_args = {}
 
-            if self.pythran_dir is not None:
-                ext_compile_flags.extend(['-I',self.pythran_dir,'-DENABLE_PYTHON_MODULE','-std=c++11','-D__PYTHRAN__=%d' % sys.version_info.major,'-Wno-cpp'])
-
             related_files = self.related_files(test_directory, module)
             self.copy_files(test_directory, workdir, related_files)
 
@@ -905,6 +914,10 @@ class CythonCompileTestCase(unittest.TestCase):
                 with open_source_file(pyx_path) as f:
                     DistutilsInfo(f).apply(extension)
 
+            if self.pythran_dir:
+                from Cython.Build.Dependencies import update_pythran_extension
+                update_pythran_extension(extension)
+
             for matcher, fixer in list(EXT_EXTRAS.items()):
                 if isinstance(matcher, str):
                     # lazy init
@@ -1334,7 +1347,7 @@ class CythonPyregrTestCase(CythonRunTestCase):
         run_forked_test(result, run_test, self.shortDescription(), self.fork)
 
 
-include_debugger = IS_CPYTHON and sys.version_info[:2] > (2, 5)
+include_debugger = IS_CPYTHON
 
 
 def collect_unittests(path, module_prefix, suite, selectors, exclude_selectors):
@@ -1831,6 +1844,9 @@ def main():
     parser.add_option("--exit-ok", dest="exit_ok", default=False,
                       action="store_true",
                       help="exit without error code even on test failures")
+    parser.add_option("--failfast", dest="failfast", default=False,
+                      action="store_true",
+                      help="stop on first failure or error")
     parser.add_option("--root-dir", dest="root_dir", default=os.path.join(DISTDIR, 'tests'),
                       help="working directory")
     parser.add_option("--work-dir", dest="work_dir", default=os.path.join(os.getcwd(), 'TEST_TMP'),
@@ -1856,25 +1872,6 @@ def main():
 
     if options.with_cython and sys.version_info[0] >= 3:
         sys.path.insert(0, options.cython_dir)
-        if sys.version_info[:2] == (3, 2):
-            try:
-                # try if Cython is installed in a Py3 version
-                import Cython.Compiler.Main
-            except Exception:
-                # back out anything the import process loaded, then
-                # 2to3 the Cython sources to make them re-importable
-                cy_modules = [ name for name in sys.modules
-                               if name == 'Cython' or name.startswith('Cython.') ]
-                for name in cy_modules:
-                    del sys.modules[name]
-                # hasn't been refactored yet - do it now
-                global CY3_DIR
-                CY3_DIR = cy3_dir = os.path.join(WORKDIR, 'Cy3')
-                refactor_for_py3(DISTDIR, cy3_dir)
-
-    if options.watermark:
-        import Cython.Compiler.Version
-        Cython.Compiler.Version.watermark = options.watermark
 
     WITH_CYTHON = options.with_cython
 
@@ -1893,20 +1890,6 @@ def main():
     if options.xml_output_dir:
         shutil.rmtree(options.xml_output_dir, ignore_errors=True)
 
-    if WITH_CYTHON:
-        global CompilationOptions, pyrex_default_options, cython_compile
-        from Cython.Compiler.Main import \
-            CompilationOptions, \
-            default_options as pyrex_default_options, \
-            compile as cython_compile
-        from Cython.Compiler import Errors
-        Errors.LEVEL = 0 # show all warnings
-        from Cython.Compiler import Options
-        Options.generate_cleanup_code = 3   # complete cleanup code
-        from Cython.Compiler import DebugFlags
-        DebugFlags.debug_temp_code_comments = 1
-        pyrex_default_options['formal_grammar'] = options.use_formal_grammar
-
     if options.shard_count > 1 and options.shard_num == -1:
         import multiprocessing
         pool = multiprocessing.Pool(options.shard_count)
@@ -1937,17 +1920,39 @@ def main():
         sys.exit(return_code)
 
 
+def configure_cython(options):
+    global CompilationOptions, pyrex_default_options, cython_compile
+    from Cython.Compiler.Main import \
+        CompilationOptions, \
+        default_options as pyrex_default_options, \
+        compile as cython_compile
+    from Cython.Compiler import Errors
+    Errors.LEVEL = 0  # show all warnings
+    from Cython.Compiler import Options
+    Options.generate_cleanup_code = 3  # complete cleanup code
+    from Cython.Compiler import DebugFlags
+    DebugFlags.debug_temp_code_comments = 1
+    pyrex_default_options['formal_grammar'] = options.use_formal_grammar
+    if options.watermark:
+        import Cython.Compiler.Version
+        Cython.Compiler.Version.watermark = options.watermark
+
+
 def runtests_callback(args):
     options, cmd_args, shard_num = args
     options.shard_num = shard_num
     return runtests(options, cmd_args)
 
+
 def runtests(options, cmd_args, coverage=None):
 
     WITH_CYTHON = options.with_cython
     ROOTDIR = os.path.abspath(options.root_dir)
     WORKDIR = os.path.abspath(options.work_dir)
 
+    if WITH_CYTHON:
+        configure_cython(options)
+
     xml_output_dir = options.xml_output_dir
     if options.shard_num > -1:
         WORKDIR = os.path.join(WORKDIR, str(options.shard_num))
@@ -1982,6 +1987,11 @@ def runtests(options, cmd_args, coverage=None):
             compiler_default_options['gdb_debug'] = True
             compiler_default_options['output_dir'] = os.getcwd()
 
+    if IS_PYPY:
+        if options.with_refnanny:
+            sys.stderr.write("Disabling refnanny in PyPy\n")
+            options.with_refnanny = False
+
     if options.with_refnanny:
         from pyximport.pyxbuild import pyx_to_dll
         libpath = pyx_to_dll(os.path.join("Cython", "Runtime", "refnanny.pyx"),
@@ -2028,6 +2038,7 @@ def runtests(options, cmd_args, coverage=None):
         exclude_selectors.append(RegExSelector('IPython'))
 
     try:
+        raise ImportError("Jedi typer is currently broken, see GH#1845")
         import jedi
         if not ([0, 9] <= list(map(int, re.findall('[0-9]+', jedi.__version__ or '0')))):
             raise ImportError
@@ -2057,9 +2068,6 @@ def runtests(options, cmd_args, coverage=None):
             for bugs_file_name, condition in bug_files if condition
         ]
 
-    if sys.platform in ['win32', 'cygwin'] and sys.version_info < (2,6):
-        exclude_selectors += [ lambda x: x == "run.specialfloat" ]
-
     global COMPILER
     if options.compiler:
         COMPILER = options.compiler
@@ -2120,7 +2128,7 @@ def runtests(options, cmd_args, coverage=None):
                                     options.cython_only, languages, test_bugs,
                                     options.fork, sys.version_info[0],
                                     options.test_determinism,
-                                    common_utility_dir, options.pythran_dir)
+                                    common_utility_dir)
             sys.stderr.write("Including CPython regression tests in %s\n" % sys_pyregr_dir)
             test_suite.addTest(filetests.handle_directory(sys_pyregr_dir, 'pyregr'))
 
@@ -2131,14 +2139,27 @@ def runtests(options, cmd_args, coverage=None):
             except OSError: pass  # concurrency issue?
         test_runner = XMLTestRunner(output=xml_output_dir,
                                     verbose=options.verbosity > 0)
+        if options.failfast:
+            sys.stderr.write("--failfast not supported with XML runner\n")
     else:
-        test_runner = unittest.TextTestRunner(verbosity=options.verbosity)
+        text_runner_options = {}
+        if options.failfast:
+            if sys.version_info < (2, 7):
+                sys.stderr.write("--failfast not supported with Python < 2.7\n")
+            else:
+                text_runner_options['failfast'] = True
+        test_runner = unittest.TextTestRunner(verbosity=options.verbosity, **text_runner_options)
 
     if options.pyximport_py:
         from pyximport import pyximport
         pyximport.install(pyimport=True, build_dir=os.path.join(WORKDIR, '_pyximport'),
                           load_py_module_on_import_failure=True, inplace=True)
 
+    try:
+        gc.set_debug(gc.DEBUG_UNCOLLECTABLE)
+    except AttributeError:
+        pass  # not available on PyPy
+
     result = test_runner.run(test_suite)
 
     if common_utility_dir and options.shard_num < 0 and options.cleanup_workdir:
